@@ -48,11 +48,12 @@
 #include <sys/types.h>
 
 /* for X11 VidMode stuff */
-#ifndef WIN32GDI
+#ifndef _WIN32
 # include <X11/Xos.h>
 # include <X11/Xlib.h>
 # include <X11/Xutil.h>
 # include <X11/extensions/xf86vmode.h>
+# include <X11/extensions/Xrandr.h>
 # ifdef FGLRX
 #  include <fglrx_gamma.h>
 # endif
@@ -76,7 +77,7 @@
 # define MAX_TABLE_SIZE   2e10
 #endif
 
-#ifdef WIN32GDI
+#ifdef _WIN32
 # define u_int16_t  WORD
 #endif
 
@@ -86,6 +87,7 @@ void error (char *fmt, ...), warning (char *fmt, ...), message(char *fmt, ...);
 #if 1
 # define BE_INT(a)    ((a)[3]+((a)[2]<<8)+((a)[1]<<16) +((a)[0]<<24))
 # define BE_SHORT(a)  ((a)[1]+((a)[0]<<8))
+# define ROUND(a)     ((a)+0.5)
 #else
 # warning "big endian is NOT TESTED"
 # define BE_INT(a)    (a)
@@ -119,9 +121,10 @@ usage (void)
   fprintf (stdout, "     or xcalib [-options] -alter\n");
   fprintf (stdout, "\n");
   fprintf (stdout, "where the available options are:\n");
-#ifndef WIN32GDI
+#ifndef _WIN32
   fprintf (stdout, "    -display <host:dpy>     or -d\n");
   fprintf (stdout, "    -screen <screen-#>      or -s\n");
+  fprintf (stdout, "    -output <output-#>      or -o\n");
 #else
   fprintf (stdout, "    -screen <monitor-#>     or -s\n");
 #endif
@@ -149,7 +152,7 @@ usage (void)
   fprintf (stdout,
 	   "last parameter must be an ICC profile containing a vcgt-tag\n");
   fprintf (stdout, "\n");
-#ifndef WIN32GDI 
+#ifndef _WIN32 
   fprintf (stdout, "Example: ./xcalib -d :0 -s 0 -v bluish.icc\n");
 #else
   fprintf (stdout, "Example: ./xcalib -v bluish.icc\n");
@@ -161,7 +164,7 @@ usage (void)
   exit (0);
 }
 
-#ifdef WIN32GDI
+#ifdef _WIN32
 /* Win32 monitor enumeration - code by gl.tter ( http://gl.tter.org ) */
 static unsigned int monitorSearchIndex = 0;
 static HDC monitorDC = 0;
@@ -211,6 +214,34 @@ HDC FindMonitor(int index)
 
 #endif
 
+float        LinInterpolateRampU16   ( unsigned short    * ramp,
+                                       int                 ramp_size,
+                                       float               pos )
+{
+  unsigned short val1, val2;
+  float start, dist, result;
+  
+  if(!ramp)
+    return 0.0;
+  
+  if(pos < 0)
+    return ramp[0];
+    
+  if(pos > ramp_size-1)
+    return ramp[ramp_size-1];
+  
+  dist = modff( pos, &start );
+  val1 = ramp[(int)start];
+  val2 = ramp[(int)start+1];
+    
+  result = val2 - val1;
+  result *= dist;
+  result += val1;
+    
+  return result;
+}
+
+
 /*
  * FUNCTION read_vcgt_internal
  *
@@ -483,9 +514,9 @@ read_vcgt_internal(const char * filename, u_int16_t * rRamp, u_int16_t * gRamp,
           for(j=0; j<numEntries; j++) {
             for(i=0; i<ratio; i++)
             {
-              rRamp[j*ratio+i] = (redRamp[j]*(ratio-i) + redRamp[j+1]*(i)) / (ratio);
-              gRamp[j*ratio+i] = (greenRamp[j]*(ratio-i) + greenRamp[j+1]*(i)) / (ratio);
-              bRamp[j*ratio+i] = (blueRamp[j]*(ratio-i) + blueRamp[j+1]*(i)) / (ratio);
+              rRamp[j*ratio+i] = (int)LinInterpolateRampU16( redRamp, numEntries, (j*ratio+i)*(double)(numEntries-1)/(double)(nEntries-1));
+              gRamp[j*ratio+i] = (int)LinInterpolateRampU16( greenRamp, numEntries, (j*ratio+i)*(double)(numEntries-1)/(double)(nEntries-1));
+              bRamp[j*ratio+i] = (int)LinInterpolateRampU16( blueRamp, numEntries, (j*ratio+i)*(double)(numEntries-1)/(double)(nEntries-1));
             }
           }
         }
@@ -504,15 +535,9 @@ read_vcgt_internal(const char * filename, u_int16_t * rRamp, u_int16_t * gRamp,
 int
 main (int argc, char *argv[])
 {
-  int fa, nfa;			/* argument we're looking at */
   char in_name[256] = { '\000' };
   char tag_name[40] = { '\000' };
-  int verb = 2;
-  int search = 0;
-  int ecount = 1;		/* Embedded count */
-  int offset = 0;		/* Offset to read profile from */
   int found;
-  int rv = 0;
   u_int16_t *r_ramp = NULL, *g_ramp = NULL, *b_ramp = NULL;
   int i;
   int clear = 0;
@@ -526,14 +551,17 @@ main (int argc, char *argv[])
   unsigned int r_res, g_res, b_res;
   int screen = -1;
 
-  unsigned int ramp_size = 256;
-  unsigned int ramp_scaling;
+#ifdef FGLRX
+  unsigned
+#endif
+           int ramp_size = 256;
 
-#ifndef WIN32GDI
+#ifndef _WIN32
   /* X11 */
   XF86VidModeGamma gamma;
   Display *dpy = NULL;
   char *displayname = NULL;
+  int xoutput = 0;
 #ifdef FGLRX
   int controller = -1;
   FGLRX_X11Gamma_C16native fglrx_gammaramps;
@@ -553,7 +581,7 @@ main (int argc, char *argv[])
   xcalib_state.verbose = 0;
 
   /* begin program part */
-#ifdef WIN32GDI
+#ifdef _WIN32
   for(i=0; i< ramp_size; i++) {
     winGammaRamp.Red[i] = i << 8;
     winGammaRamp.Blue[i] = i << 8;
@@ -563,7 +591,7 @@ main (int argc, char *argv[])
 
   /* command line parsing */
   
-#ifndef WIN32GDI
+#ifndef _WIN32
   if (argc < 2)
     usage ();
 #endif
@@ -584,7 +612,7 @@ main (int argc, char *argv[])
         fprintf(stdout, "xcalib " XCALIB_VERSION "\n");
         exit (0);
     }
-#ifndef WIN32GDI
+#ifndef _WIN32
     /* X11 display */
     if (!strcmp (argv[i], "-d") || !strcmp (argv[i], "-display")) {
       if (++i >= argc)
@@ -600,6 +628,15 @@ main (int argc, char *argv[])
       screen = atoi (argv[i]);
       continue;
     }
+#ifndef _WIN32
+    /* X11 output */
+    if (!strcmp (argv[i], "-o") || !strcmp (argv[i], "-output")) {
+      if (++i >= argc)
+        usage ();
+        xoutput = atoi (argv[i]);
+        continue;
+    }
+#endif
 #ifdef FGLRX
     /* ATI controller index (for FGLRX only) */
     if (!strcmp (argv[i], "-x") || !strcmp (argv[i], "-controller")) {
@@ -809,7 +846,7 @@ main (int argc, char *argv[])
     }
   }
 
-#ifdef WIN32GDI
+#ifdef _WIN32
   if ((!clear || !alter) && (in_name[0] == '\0')) {
     hDc = FindMonitor(screen);
     win_profile_len = MAX_PATH;
@@ -827,7 +864,7 @@ main (int argc, char *argv[])
   }
 #endif
 
-#ifndef WIN32GDI
+#ifndef _WIN32
   /* X11 initializing */
   if ((dpy = XOpenDisplay (displayname)) == NULL) {
     if(!donothing)
@@ -838,13 +875,61 @@ main (int argc, char *argv[])
   else if (screen == -1)
     screen = DefaultScreen (dpy);
 
+  int xrr_version = -1;
+  int crtc = 0;
+  int major_versionp = 0;
+  int minor_versionp = 0;
+  int n = 0;
+  Window root = RootWindow(dpy, screen);
+
+  XRRQueryVersion( dpy, &major_versionp, &minor_versionp );
+  xrr_version = major_versionp*100 + minor_versionp;
+
+  if(xrr_version >= 102)
+  {                           
+    XRRScreenResources * res = XRRGetScreenResources( dpy, root );
+    int ncrtc = 0;
+
+    n = res->noutput;
+    for( i = 0; i < n; ++i )
+    {
+      RROutput output = res->outputs[i];
+      XRROutputInfo * output_info = XRRGetOutputInfo( dpy, res,
+                                                        output);
+      if(output_info->crtc)
+        if(ncrtc++ == xoutput)
+        {
+          crtc = output_info->crtc;
+          ramp_size = XRRGetCrtcGammaSize( dpy, crtc );
+          message ("XRandR output:      \t%s\n", output_info->name);
+        }
+
+      XRRFreeOutputInfo( output_info ); output_info = 0;
+    }
+    //XRRFreeScreenResources(res); res = 0;
+  }
+
   /* clean gamma table if option set */
   gamma.red = 1.0;
   gamma.green = 1.0;
   gamma.blue = 1.0;
   if (clear) {
 #ifndef FGLRX
-    if (!XF86VidModeSetGamma (dpy, screen, &gamma)) {
+    if(xrr_version >= 102)
+    {
+      XRRCrtcGamma * gamma = XRRAllocGamma (ramp_size);
+      if(!gamma)
+        warning ("Unable to clear screen gamma");
+      else
+      {
+        for(i=0; i < ramp_size; ++i)
+          gamma->red[i] = gamma->green[i] = gamma->blue[i] = i * 65535 / ramp_size;
+        XRRSetCrtcGamma (dpy, crtc, gamma);
+        XRRFreeGamma (gamma);
+      }
+    } else
+    if (!XF86VidModeSetGamma (dpy, screen, &gamma))
+    {
 #else
     for(i = 0; i < 256; i++) {
       fglrx_gammaramps.RGamma[i] = i << 2;
@@ -863,7 +948,7 @@ main (int argc, char *argv[])
   if(!donothing)
   {
 #ifndef FGLRX
-    if (!XF86VidModeGetGammaRampSize (dpy, screen, &ramp_size)) {
+    if (xrr_version < 102 && !XF86VidModeGetGammaRampSize (dpy, screen, &ramp_size)) {
 #else
     if (!FGLRX_X11GetGammaRampSize(dpy, screen, &ramp_size)) {
 #endif
@@ -876,7 +961,7 @@ main (int argc, char *argv[])
       }
     }
   }
-#else /* WIN32GDI */
+#else /* _WIN32 */
   if(!donothing) {
     if(!hDc)
       hDc = FindMonitor(screen);
@@ -918,7 +1003,7 @@ main (int argc, char *argv[])
     if( (i = read_vcgt_internal(in_name, r_ramp, g_ramp, b_ramp, ramp_size)) <= 0) {
       if(i<0)
         warning ("Unable to read file '%s'", in_name);
-      if(i=0)
+      if(i == 0)
         warning ("No calibration data in ICC profile '%s' found", in_name);
       free(r_ramp);
       free(g_ramp);
@@ -926,8 +1011,20 @@ main (int argc, char *argv[])
       exit(0);
     }
   } else {
-#ifndef WIN32GDI
-    if (!XF86VidModeGetGammaRamp (dpy, screen, ramp_size, r_ramp, g_ramp, b_ramp))
+#ifndef _WIN32
+    if (xrr_version >= 102)
+    {
+      XRRCrtcGamma * gamma = 0;
+      if((gamma = XRRGetCrtcGamma(dpy, crtc)) != 0 )
+        warning ("Unable to get display calibration");
+
+      for (i = 0; i < ramp_size; i++) {
+        r_ramp[i] = gamma->red[i];
+        g_ramp[i] = gamma->green[i];
+        b_ramp[i] = gamma->blue[i];
+      }
+    }
+    else if (!XF86VidModeGetGammaRamp (dpy, screen, ramp_size, r_ramp, g_ramp, b_ramp))
       warning ("Unable to get display calibration");
 #else
     if (!GetDeviceGammaRamp(hDc, &winGammaRamp))
@@ -946,7 +1043,6 @@ main (int argc, char *argv[])
     float redContrast = 100.0;
     float redMin = 0.0;
     float redMax = 1.0;
-    float redGamma = 1.0;
 
     redMin = (double)r_ramp[0] / 65535.0;
     redMax = (double)r_ramp[ramp_size - 1] / 65535.0;
@@ -960,7 +1056,6 @@ main (int argc, char *argv[])
     float greenContrast = 100.0;
     float greenMin = 0.0;
     float greenMax = 1.0;
-    float greenGamma = 1.0;
 
     greenMin = (double)g_ramp[0] / 65535.0;
     greenMax = (double)g_ramp[ramp_size - 1] / 65535.0;
@@ -974,7 +1069,6 @@ main (int argc, char *argv[])
     float blueContrast = 100.0;
     float blueMin = 0.0;
     float blueMax = 1.0;
-    float blueGamma = 1.0;
 
     blueMin = (double)b_ramp[0] / 65535.0;
     blueMax = (double)b_ramp[ramp_size - 1] / 65535.0;
@@ -1006,14 +1100,14 @@ main (int argc, char *argv[])
   }
 
   if(!invert) {
-    /* ramps should be monotonic - otherwise content is nonsense! */
+    /* ramps should be increasing - otherwise content is nonsense! */
     for (i = 0; i < ramp_size - 1; i++) {
       if (r_ramp[i + 1] < r_ramp[i])
-        warning ("red gamma table not monotonic");
+        warning ("red gamma table not increasing");
       if (g_ramp[i + 1] < g_ramp[i])
-        warning ("green gamma table not monotonic");
+        warning ("green gamma table not increasing");
       if (b_ramp[i + 1] < b_ramp[i])
-        warning ("blue gamma table not monotonic");
+        warning ("blue gamma table not increasing");
     }
   } else {
     for (i = 0; i < ramp_size; i++) {
@@ -1056,9 +1150,9 @@ main (int argc, char *argv[])
       }
       tmpRampVal = b_ramp[i];
     }
-    fprintf(stdout, "R: %d\tG: %d\t B: %d\t colors lost\n", ramp_size - r_res, ramp_size - g_res, ramp_size - b_res );
+    fprintf(stdout, "R: %d  G: %d  B: %d  colors lost\n", ramp_size - r_res, ramp_size - g_res, ramp_size - b_res );
   }
-#ifdef WIN32GDI
+#ifdef _WIN32
   for (i = 0; i < ramp_size; i++) {
     winGammaRamp.Red[i] = r_ramp[i];
     winGammaRamp.Green[i] = g_ramp[i];
@@ -1069,11 +1163,11 @@ main (int argc, char *argv[])
  
   if(printramps)
     for(i=0; i<ramp_size; i++)
-      fprintf(stdout,"%x %x %x\n", r_ramp[i], g_ramp[i], b_ramp[i]);
+      fprintf(stdout,"%d %d %d\n", r_ramp[i], g_ramp[i], b_ramp[i]);
 
   if(!donothing) {
     /* write gamma ramp to X-server */
-#ifndef WIN32GDI
+#ifndef _WIN32
 # ifdef FGLRX
     for(i = 0; i < ramp_size; i++) {
       fglrx_gammaramps.RGamma[i] = r_ramp[i] >> 6;
@@ -1082,6 +1176,23 @@ main (int argc, char *argv[])
     }
     if (!FGLRX_X11SetGammaRamp_C16native_1024(dpy, screen, controller, ramp_size, &fglrx_gammaramps))
 # else
+    if(xrr_version >= 102)
+    {
+      XRRCrtcGamma * gamma = XRRAllocGamma (ramp_size);
+      if(!gamma)
+        warning ("Unable to calibrate display");
+      else
+      {
+        for(i=0; i < ramp_size; ++i)
+        {
+          gamma->red[i] = r_ramp[i];
+          gamma->green[i] = g_ramp[i];
+          gamma->blue[i] = b_ramp[i];
+        }
+        XRRSetCrtcGamma (dpy, crtc, gamma);
+        XRRFreeGamma (gamma);
+      }
+    } else
     if (!XF86VidModeSetGammaRamp (dpy, screen, ramp_size, r_ramp, g_ramp, b_ramp))
 # endif
 #else
@@ -1097,7 +1208,7 @@ main (int argc, char *argv[])
   free(b_ramp);
 
 cleanupX:
-#ifndef WIN32GDI
+#ifndef _WIN32
   if(dpy)
     if(!donothing)
       XCloseDisplay (dpy);
